import { v4 as uuidv4 } from 'uuid'
import dotenv from 'dotenv';
import { TransportEvent, ConnectionManager as ConnectionManagerInterface, ConnectionSet, Transport, TransportService } from "../interface/connector.interface"
import { TransmitterConnectionAdapter } from './connector.transmitter'
import { ReceiverConnectionAdapter } from './connector.receiver'
import { TransmissionEvent } from '../interface/transport.interface'
import { Subject } from 'rxjs'
import { WebsocketTransportService } from '../transport/websocket'
import { RequestResponseConnectionAdapter } from './connector.request.response'
import { HttpTransportService } from '../transport/http'
/* now, this connection manager, after setting up the servers, will also need to start assigning receivers so to speak.
From Provider's perspective: When instantiating a new server, there will be no doubt multiple clients connecting in. So that means, the connector instances that will be used from
the transmission side, must already be configured and locked to their respective target so to speak. So for one client, there would only be one transmitter and receiver connector instances
as well as request response. Let's go along with this line for now. */
export class ConnectionManager implements ConnectionManagerInterface {
    transportService!: TransportService;
    transmissionEvent: Subject<TransmissionEvent> = new Subject()
    transportEvent: Subject<TransportEvent> = new Subject() // Every event goes through this boy, and it will be expose across differnet components
    transmissionSet: ConnectionSet[] = []

    constructor(messageTransmissionEvent: Subject<TransmissionEvent>) {
        this.transmissionEvent = messageTransmissionEvent 
        // logic here
        this.setUpServer(process.env.PORT as unknown as number, this.getTransportService(process.env.Transport as unknown as Transport))

        // will also have information on events happening on transmission side
        this.transmissionEvent.subscribe((event: TransmissionEvent) => {
            console.log(`TransmissionEvent: `, event)
        })
        // Everything that happens will go through here. Can also plug in logging service here at a later time
        this.transportEvent.subscribe((event: TransportEvent) => {
            console.log(`TranpsortEvent: `, event)
        })
    }

    /* All three methods that return their respective instances will have it's activities reported through transportEvent */
    getTransmitterConnectionAdapter(): TransmitterConnectionAdapter {
        if (this.transportService) {
            let adapter: TransmitterConnectionAdapter = new TransmitterConnectionAdapter(this.transportService)
            return adapter
        } else {
            throw new Error(`Transmitter Transport NOT initialized.`)
        }
    }

    getReceiverConnectionAdapter(): ReceiverConnectionAdapter {
        if (this.transportService) {
            let adapter: ReceiverConnectionAdapter = new ReceiverConnectionAdapter(this.transportService)
            return adapter
        } else {
            throw new Error(`Receiver Transport NOT initialized.`)
        }
    }

    getRequestResponseConnectionAdapter(transmitterAdapter: TransmitterConnectionAdapter, receiverAdapter: ReceiverConnectionAdapter): RequestResponseConnectionAdapter {
        if (this.transportService) {
            let adapter: RequestResponseConnectionAdapter = new RequestResponseConnectionAdapter(transmitterAdapter, receiverAdapter)
            return adapter
        } else {
            throw new Error(`Request Response Transport NOT initialized.`)
        }
    }

    getTransportService(transportType: Transport): any {
        console.log(`Instanting ${transportType} tranport service to be used.`)
        if (transportType == Transport.Websocket) {
            this.transportService = new WebsocketTransportService()
            this.transportEvent.next({
                id: uuidv4(),
                event: `New Transport`,
                data: {
                    message: `Setting up Websocket Transport Service`
                }
            })
        }
        if (transportType == Transport.Http) {
            this.transportService = new HttpTransportService()
            this.transportEvent.next({
                id: uuidv4(),
                event: `New Transport`,
                data: {
                    message: `Setting up Http Transport Service`
                }
            })
        }
    }

    // Server to be set up as well as acquiring client information if needed. Like in the case for grpc and socket. Http not required.
    private setUpServer(port: number, transportService: TransportService): void {
        if ((transportService as TransportService).getInfo().transport == Transport.Websocket) {
            (transportService as WebsocketTransportService).startServer(port);
            // this doesn't mean all the responses come through here, but rather to record and assign new instances based on the new clinets connected
            (transportService as WebsocketTransportService).getTransportEvent().subscribe(this.transportEvent)
        }

        if ((transportService as TransportService).getInfo().transport == Transport.Http) {
            (transportService as HttpTransportService).startServer(port)
            //  To be Enhanced at a later time
        }
    }
}